(*
  All the code presented in this unit is Copyright Peter D Johnson
  www.delphidabbler.com, (c) 2004.

  The code may be used in any way providing it is not sold or passed off as
  the work of another person.

  The code is used at the user's own risk. No guarantee or warranty is provided.
*)

unit USingleInst;

interface

uses
   Windows, Controls, Messages, Dialogs;

type
   TSingleInstParamHandler = procedure(const Param: string) of object;

type
   TSingleInstClass = class of TSingleInst;

   TSingleInst = class (TObject)
   private
      fOnProcessParam: TSingleInstParamHandler; // event handler
      fEnsureRestoreMsg: UINT;  // unique message for ensuring wdw restoration
   protected
      function WdwClassName: string; virtual;
      {Returns 'SingleInst.MainWdw': override in descendant class to return
      unique window class name}
      function WaterMark: DWORD; virtual;
      {Returns 0: override in sub class to set "unique" value}
      function FindDuplicateMainWdw: HWND; virtual;
      {Only override if want to use different method of finding window}
      function SendParamsToPrevInst(Wdw: HWND): boolean; virtual;
      {Only override if want to use different encoding of data or to send data
      other than command line params. Also override WMCopyData.}
      function SwitchToPrevInst(Wdw: HWND): boolean;
      procedure EnsureRestore(var Msg: TMessage); dynamic;
      {Override to change window restoration behaviour}
      procedure WMCopyData(var Msg: TMessage); dynamic;
      {Only override if SendParamsToPrevInst overridden}
   public
      constructor Create;
      {Do not call directly: singleton object}
      procedure CreateParams(var Params: TCreateParams);
      {ENTRY POINT: override main form's CreateParams method and call from
      there}
      function HandleMessages(var Msg: TMessage): boolean;
      {ENTRY POINT: override main form's WndProc method and call from there. If
      returns false call WndProc's inherited method}
      function CanStartApp: boolean;
      {ENTRY POINT: call from project file to determine if to start application}
      property OnProcessParam: TSingleInstParamHandler
         Read fOnProcessParam Write fOnProcessParam;
      {Handle this event in main form to process parameters passed in
      WM_COPYDATA}
   end;

function SingleInst: TSingleInst;
{Returns singleton TSingleInst object: object created first time called}

procedure RegisterSingleInstClass(Cls: TSingleInstClass);
  {Registers a sub class of TSingleInst to use as SingleInst singleton: call
  this function from the initialization section of the unit where the class is
  defined - must be called before SingleInst}

implementation

uses
   SysUtils, Forms, u_Main, u_ShLowUtils;

var
   // Globals storing SingleInst singleton and class of SingleInst
   gSingleInst: TSingleInst = nil;
   gSingleInstClass: TSingleInstClass = nil;

function SingleInst: TSingleInst;
begin
   if not Assigned(gSingleInst) then
   begin
      if Assigned(gSingleInstClass) then
         gSingleInst := gSingleInstClass.Create
      else
         gSingleInst := TSingleInst.Create;
   end;
   Result := gSingleInst;
end;

procedure RegisterSingleInstClass(Cls: TSingleInstClass);
begin
   gSingleInstClass := Cls;
end;

{ TSingleInst }

function TSingleInst.CanStartApp: boolean;
var
   Wdw: HWND;
begin
   Wdw := FindDuplicateMainWdw;
   if Wdw = 0 then
      Result := True
   else
      Result := not SwitchToPrevInst(Wdw);
end;

constructor TSingleInst.Create;
begin
   inherited;
   fEnsureRestoreMsg := RegisterWindowMessage('USINGLEINST_ENSURERESTORE');
end;

procedure TSingleInst.CreateParams(var Params: TCreateParams);
begin
   inherited;
   StrPLCopy(Params.WinClassName, WdwClassName, SizeOf(Params.WinClassName) - 1);
end;

procedure TSingleInst.EnsureRestore(var Msg: TMessage);
begin
   if IsIconic(Application.Handle) then
      Application.Restore;
   if Assigned(Application.MainForm) and not Application.MainForm.Visible then
      Application.MainForm.Visible := True;
   Application.BringToFront;
end;

function TSingleInst.FindDuplicateMainWdw: HWND;
begin
   Result := FindWindow(PChar(WdwClassName), nil);
end;

function TSingleInst.HandleMessages(var Msg: TMessage): boolean;
var
   filename: string;
begin
   if Msg.Msg = WM_COPYDATA then
   begin
      // f.schneider v3.5.0.36 [2883]
      // die Methode HandleParam aus Frm_Main wurde hier integriert,
      // damit sowohl die Parameterbergabe an SingleInstance funktioniert
      // als auch die SendMessage zum Viewer fr den Usernamen abzufragen
      if TWMCopyData(Msg).CopyDataStruct.dwData = $DDED then
      begin
         PData := TWMCopyData(Msg).CopyDataStruct.lpData;
         while PData^ <> #0 do
         begin
            Param := StrPas(PData);
            GGfLogEvnt('TSingleInst.HandleMessages Param=' + Param);

            if Pos('-f:',Param) > 0 then
            begin
               filename := Copy(Param,4, Length(Param) -3);
               frm_Main.loadfile(filename);
            end
            else
            if Pos('-refresh', Param) > 0 then
            begin
               frm_Main.lgr_Arc_click();
            end
            else
            if sGusrname = '' then
               sGusrname := Param;
            gbreceivedviewermessage := true;

            Inc(PData, Length(Param) + 1);
         end;
         Msg.Result := 1;
      end
      else
      begin
         Msg.Result := 0;
         application.ProcessMessages;
         WMCopyData(Msg);
      end;

      Result := True;
   end
   else
   if Msg.Msg = fEnsureRestoreMsg then
   begin
      EnsureRestore(Msg);
      Result := True;
   end
   else
      Result := False;
end;

function TSingleInst.SendParamsToPrevInst(Wdw: HWND): boolean;
var
   CopyData: TCopyDataStruct;
   I: integer;
   DataSize: integer;
   Data: PChar;
   PData: PChar;
begin
   DataSize := 0;
   for I := 1 to ParamCount do
      Inc(DataSize, Length(ParamStr(I)) + 1);
   Inc(DataSize);
   Data := StrAlloc(DataSize);
   try
      PData := Data;
      for I := 1 to ParamCount do
      begin
         StrPCopy(PData, ParamStr(I));
         Inc(PData, Length(ParamStr(I)) + 1);
      end;
      PData^ := #0;
      CopyData.lpData := Data;
      CopyData.cbData := DataSize;
      CopyData.dwData := WaterMark;
      Result := SendMessage(Wdw, WM_COPYDATA, 0, LPARAM(@CopyData)) = 1;
   finally
      StrDispose(Data);
   end;
end;

function TSingleInst.SwitchToPrevInst(Wdw: HWND): boolean;
begin
   Assert(Wdw <> 0);
   if ParamCount > 0 then
      Result := SendParamsToPrevInst(Wdw)
   else
      Result := True;
   if Result then
      SendMessage(Wdw, fEnsureRestoreMsg, 0, 0);
end;

function TSingleInst.WaterMark: DWORD;
begin
   Result := $DDED;
end;

function TSingleInst.WdwClassName: string;
begin
   Result := 'bitfarm.Manuscan.V3';
end;

procedure TSingleInst.WMCopyData(var Msg: TMessage);
var
   PData: PChar;
   Param: string;
begin
   if TWMCopyData(Msg).CopyDataStruct.dwData = WaterMark then
   begin
      PData := TWMCopyData(Msg).CopyDataStruct.lpData;
      while PData^ <> #0 do
      begin
         Param := StrPas(PData);
         if Assigned(fOnProcessParam) then
            fOnProcessParam(Param);
         Inc(PData, Length(Param) + 1);
      end;
      Msg.Result := 1;
   end
   else
      Msg.Result := 0;
   application.ProcessMessages;
end;

initialization
   RegisterSingleInstClass(TSingleInst);


finalization

   gSingleInst.Free;

end.
