unit u_ShRecTab;

(*
  ------------------------------------------------------------------------------
*)
interface
(*
  ------------------------------------------------------------------------------
*)

uses
  Classes;

type
  clRecAbstr = class
  protected
    function isEqual (Rec: TObject): Boolean;  virtual; abstract;
  end;


  (*
    Tabelle fr beliebige Records

    Die Klasse fr Record-Tabellen ermglicht den Zugriff auf beliebige
    Datenstze und pat sich den aktuellen Datenmengen an.
    Der Algorithmus arbeitet mit einem sogenannten 'aktuellen Satz', der
    fr die Methoden als Parameter dient. Dadurch knnen die meisten
    Methoden parameterlos bleiben, was die Wiederverwendbarkeit in
    unterschiedlichen Umgebungen ermglicht.

    Die Methoden 'add', 'insert' und 'setRec' fgen einen Satz in die Tabelle
    ein. Um eine Typsicherheit zu erzwingen werden sie als 'protected'
    dekalriert. Eine konkrete Implementierung von 'clRecTab' mu diese
    Methoden mit den korrekten Datentypen ausstatten.
  *)
  clRecTab = class
  private
    FList : TList;                     //  Tabelle der Datenstze
    FRecNo: integer;                   //  Index des aktuellen Satzes
    FEOF  : Boolean;                   //  wenn in einer for next Schleife der letze Datensatz erreicht wurde

  protected
    (*
      Zweck:
        Schreibt den akuellen Satz in die Tabelle
      Eingabe:
        'Rec': Satz, der an die Stelle 'FRecNo' geschrieben wird
    *)
    procedure setRec (const Rec: TObject);


    (*
      Zweck:
        Fgt einen Satz am Ende der Tabelle an
      Eingabe:
        'Rec': Der einzufgende Satz
    *)
    procedure Add (const Rec: TObject);  virtual;


    (*
      Zweck:
        Fgt einen Satz in die Tabelle ein
      Eingabe:
        'Idx': Index, an dem der Satz eingefgt wird.
               Der vorher hier stehende Satz wird, wie alle nachfolgenden
               Stze, nach hinten verschoden. Diese Stze haben
               anschlieend einen um 1 greren Index.
        'Rec': Der einzufgende Satz
    *)
    procedure Insert (      Idx: integer;
                      const Rec: TObject);


  public
    constructor Create;
    destructor  Destroy;  override;

    property EOF: Boolean Read FEOF;     	  // read-only

    (*
      Zweck:
        Liest den aktuellen Satz aus der Tabelle
      Rckgabe:
        Der aktuelle Satz, der an die Stelle 'FRecNo' steht
    *)
    function  getRec : TObject; overload;


    (*
      Zweck:
        Liest einen Satz aus der Tabelle. Schnelle Variante
      Eingabe:
        'Idx': Tabellen-Index, an dem der gewnschte Satz steht
      Rckgabe:
        Satz an der Stelle 'Idx'
    *)
    function  getRec (Idx: integer): TObject; overload;


    (*
      Zweck:
        Ermittelt den Index des aktuellen Satzes
    *)
    function  getRecNo : integer;


    (*
      Zweck:
        Ermittelt die Anzahl der gespeicherten Datenstze
    *)
    function  Count : integer;


    (*
      Zweck:
        Bestimmt den vorherigen Satz als den aktuellen Satz
    *)
    procedure Prev;


    (*
      Zweck:
        Bestimmt den folgenden Satz als den aktuellen Satz
    *)
    procedure Next;


    (*
      Zweck:
        Bestimmt den ersten Satz als den aktuellen Satz
    *)
    procedure First;


    (*
      Zweck:
        Bestimmt den letzten Satz als den aktuellen Satz
    *)
    procedure Last;


    (*
      Zweck:
        Bestimmt einen Satz als den aktuellen Satz
      Eingabe:
        'Idx': Der Satz mit diesem Index ist der aktuelle
    *)
    procedure setToRecNo (Idx: integer);


    (*
      Zweck:
        Testet, ob aktuelle Satz der erste Satz ist
    *)
    function  isFirst : boolean;


    (*
      Zweck:
        Testet, ob aktuelle Satz der letzt Satz ist
    *)
    function  isLast : boolean;


    (*
      Zweck:
        Entfernt Stze aus der Tabelle
      Eingabe:
        'startIdx': Index des ersten zu entfernenden Satzes
        'endIdx'  : Index des ersten nicht entfernten Satzes
      Hinweis:
        Nach dem Vorgang hat der Satz, der vorher an der Stelle 'EndIdx'
        stand, den Index 'StartIdx'. Alle folgenden Stze werden entsprechend
        nach vorne verschoben.
      Achtung!
        Fr alle gelschte Objekte wird die Funktion 'Free' aufgerufen!
    *)
    procedure Remove (StartIdx: integer;
                      EndIdx  : integer);


    (*
      Zweck:
        Entfernt alle Stze aus der Tabelle
    *)
    procedure Clear;


    (*
       Zweck:
         Findet den Datensatz, der als Parameter definiert wurde.
       Rckgabe:
         Den gesuchten Datensatz
    *)
    function Find (Rec: clRecAbstr): clRecAbstr;
  end;

(*
  ------------------------------------------------------------------------------
*)
implementation
(*
  ------------------------------------------------------------------------------
*)

(*  clRecTab
  *)
function clRecTab.Find (Rec: clRecAbstr): clRecAbstr;

  var
    lRec   : clRecAbstr;
    lIdx   : Integer;
    lIdxMax: Integer;

  begin
    Result  := nil;
    lIdxMax := Count -1;

    for lIdx := 0 to lIdxMax do
    begin
      lRec := clRecAbstr (getRec (lIdx));

      if Rec.isEqual (lRec) then
      begin
        Result := lRec;
        Exit;
      end;
    end;
  end;


procedure clRecTab.Add (const Rec: TObject);
  begin
    FList.Add (Rec);
  end;


constructor clRecTab.Create;
  begin
    FList  := TList.Create;
    FRecNo := 0;
    FEOF := False;
  end;


destructor clRecTab.Destroy;

  var
    xI: integer;

  begin
    for xI := 0 to Count -1 do
      getRec (xI).Free;

    FList.Free;

    inherited;
  end;


function clRecTab.Count: integer;
  begin
    Result := FList.Count;
    if FList.Count = 0 then
    	FEOF := True;
  end;


function clRecTab.getRec (Idx: integer): TObject;
  begin
    Result := FList.Items [Idx];
  end;


function clRecTab.getRec: TObject;
  begin
    Result := FList.Items [FRecNo];
  end;


function clRecTab.getRecNo: integer;
  begin
    Result := FRecNo;
  end;


procedure clRecTab.Insert (      Idx: integer;
                           const Rec: TObject);
  begin
    FList.Insert (Idx, Rec);
  end;


function clRecTab.isFirst: boolean;
  begin
    Result := FRecNo = 0;
  end;


function clRecTab.isLast: boolean;
  begin
    Result := FRecNo = FList.Count - 1;
  end;


procedure clRecTab.Next;
  begin
    if FRecNo < FList.Count - 1 then
      FRecNo := FRecNo + 1
	 else
    	FEOF := True;
  end;


procedure clRecTab.Prev;
  begin
    if FRecNo > 0 then
      FRecNo := FRecNo - 1;
    FEOF := False;
  end;


procedure clRecTab.Remove (StartIdx,
                           EndIdx  : integer);

  var
    xObj: TObject;
    xI  : integer;

  begin
    for xI := StartIdx to EndIdx do begin
      xObj := getRec (StartIdx);
      FList.Delete   (StartIdx);
      xObj .Free;
    end;
  end;


procedure clRecTab.Clear;

  var
    xObj: TObject;
    xI  : integer;

  begin
    for xI := 0 to Count-1 do begin
      xObj := getRec (xI);
      xObj.Free;
    end;

    FList.Clear;
    FEOF := False;
  end;


procedure clRecTab.setRec (const Rec: TObject);
  begin
    FList.Items [FRecNo] := Rec;
  end;


procedure clRecTab.First;
  begin
    FRecNo := 0;
    FEOF := False;
  end;


procedure clRecTab.Last;
  begin
    FRecNo := FList.Count - 1;
    FEOF := True;
  end;


procedure clRecTab.setToRecNo (Idx: integer);
  begin
    if    (Idx >= 0          )
      and (Idx <  FList.Count)
    then
      FRecNo := Idx;
    if Idx = FList.Count -1 then
    	FEOF := True;
  end;

initialization
end.


